#include "globals.h"
#include "DepthMap.h"
#include "plyio.h"
#include "meshAux.h"
#include "softRender.h"

int
main(int argc, char **argv)
{
   float resolution, xmin, xmax, ymin, ymax;
   int xdim, ydim;

   RangeGrid *rangeGrid = new RangeGrid;
   rangeGrid->readPly(argv[1]);

   Mesh *mesh = meshFromGrid(rangeGrid, 1, USE_VIEW_DOT_NORMAL, 0.1);
   mesh->writePly(argv[2]);

   mesh->computeBBox();

   resolution = 0.0005;

   xmin = mesh->bbox.fll.x;
   xmax = mesh->bbox.nur.x;
   ymin = mesh->bbox.fll.y;
   ymax = mesh->bbox.nur.y;

   xdim = int((xmax-xmin)/resolution)+1;
   ydim = int((ymax-ymin)/resolution)+1;

   Matrix4f mat;
   mat.makeIdentity();
   mat.setScale(1/resolution, 1/resolution, 1);
   mat.setTranslate(-xmin/resolution, -ymin/resolution, 0);

   mesh->applyXform(mat);
   mesh->computeBBox();

   Projector *proj = new Projector;
   proj->mat.setValue(mat);
   proj->matInv.setValue(mat.inverse());
   
   DepthMap *depthMap = new DepthMap(xdim,ydim);
   depthMap->proj = proj;
   softRender(depthMap, mesh);

   RangeGrid *resampGrid = depthMap->buildRangeGrid();
   resampGrid->writePly(argv[3]);
}

