#include <stdio.h>
#include <ply.h>
#include <stdlib.h>
//changed by lizhang
#ifdef LINUX
#include <strings.h>
#endif
#ifdef WIN32
#include <string.h>
#include <float.h>
#endif
//end of change
#include <string.h>
#include <iostream.h>
#include <limits.h>

//changed by lizhang
#ifdef WIN32
#define PATH_MAX 256
#endif
//end of change
#define MAX_VERT_PROPS 20

#include "meshAux.h"
#include "plyio.h"


struct PlyVertex {
    float x, y, z;
    float nx, ny, nz;
    uchar diff_r, diff_g, diff_b;
    float intensity;
    float confidence;
};

struct PlyFace {
    uchar nverts;
    int *vindex;
};

struct PlyRangePt {
   uchar npts;
   int *pindex;
};

static PlyProperty vert_prop_x =  
   {"x", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_y =  
  {"y", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_z =  
  {"z", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_nx =  
   {"nx", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_ny =  
  {"ny", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_nz =  
  {"nz", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_intensity =  
  {"intensity", PLY_FLOAT, PLY_FLOAT, 0, 0, PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_confidence =  
  {"confidence", PLY_FLOAT, PLY_FLOAT, 0, 0, 
   PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_diff_r =  
  {"diffuse_red", PLY_UCHAR, PLY_UCHAR, 0, 0, 
   PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_diff_g =  
  {"diffuse_green", PLY_UCHAR, PLY_UCHAR, 0, 0, 
   PLY_START_TYPE, PLY_START_TYPE, 0};
static PlyProperty vert_prop_diff_b =  
  {"diffuse_blue", PLY_UCHAR, PLY_UCHAR, 0, 0, 
   PLY_START_TYPE, PLY_START_TYPE, 0};

static PlyProperty vert_props[MAX_VERT_PROPS];


static PlyProperty face_props[] = { 
   {"vertex_indices", PLY_INT, PLY_INT, 0, 1, PLY_UCHAR, PLY_UCHAR, 0},
};


/* list of property information for a range data point */
static PlyProperty range_props[] = {
   {"vertex_indices", PLY_INT, PLY_INT, 0, 1, PLY_UCHAR, PLY_UCHAR, 0},
};


int
setupVertProps(PlyFile *ply, PlyProperty *vert_props, 
	       int hasNormals, int hasIntensity, int hasConfidence,
	       int hasColor)
{
    int nvp = 0;

    // Every vertex has x,y,z coords
    vert_props[nvp] = vert_prop_x;
    vert_props[nvp].offset = offsetof(PlyVertex, x); nvp++;
    vert_props[nvp] = vert_prop_y;
    vert_props[nvp].offset = offsetof(PlyVertex, y); nvp++;
    vert_props[nvp] = vert_prop_z;
    vert_props[nvp].offset = offsetof(PlyVertex, z); nvp++;
    
    if (hasNormals) {
	vert_props[nvp] = vert_prop_nx;
	vert_props[nvp].offset = offsetof(PlyVertex, nx); nvp++;
	vert_props[nvp] = vert_prop_ny;
	vert_props[nvp].offset = offsetof(PlyVertex, ny); nvp++;
	vert_props[nvp] = vert_prop_nz;
	vert_props[nvp].offset = offsetof(PlyVertex, nz); nvp++;
    }
    
    if (hasIntensity) {
	vert_props[nvp] = vert_prop_intensity;
	vert_props[nvp].offset = offsetof(PlyVertex, intensity); nvp++;
    }
    
    if (hasConfidence) {
	vert_props[nvp] = vert_prop_confidence;
	vert_props[nvp].offset = offsetof(PlyVertex, confidence); nvp++;
    }
    
    if (hasColor)
    {
	vert_props[nvp] = vert_prop_diff_r;
	vert_props[nvp].offset = offsetof(PlyVertex, diff_r); nvp++;
	vert_props[nvp] = vert_prop_diff_g;
	vert_props[nvp].offset = offsetof(PlyVertex, diff_g); nvp++;
	vert_props[nvp] = vert_prop_diff_b;
	vert_props[nvp].offset = offsetof(PlyVertex, diff_b); nvp++;
    }
    
    return nvp;
}


void
whichVertProps(PlyFile *ply, int& hasNormals, int& hasIntensity, 
	       int& hasConfidence, int& hasColor)
{
   hasNormals = 
      ply_is_valid_property(ply, "vertex", vert_prop_nx.name) &&
      ply_is_valid_property(ply, "vertex", vert_prop_ny.name) &&
      ply_is_valid_property(ply, "vertex", vert_prop_nz.name);
   
   hasIntensity = ply_is_valid_property
      (ply, "vertex", vert_prop_intensity.name);
   
   hasConfidence = ply_is_valid_property
      (ply, "vertex", vert_prop_confidence.name);
   
   hasColor = 
      ply_is_valid_property(ply, "vertex", vert_prop_diff_r.name) &&
      ply_is_valid_property(ply, "vertex", vert_prop_diff_g.name) &&
      ply_is_valid_property(ply, "vertex", vert_prop_diff_b.name);
}

int
setupFaceProps(PlyFile *ply, PlyProperty *face_props)
{
   int nfp = 0;
   
   face_props[nfp].offset = offsetof(PlyFace, vindex);
   face_props[nfp].count_offset = offsetof(PlyFace, nverts);  
   nfp++;

   return nfp;
}

int
setupRangeProps(PlyFile *ply, PlyProperty *range_props)
{
   int nrp = 0;
   
   range_props[nrp].offset = offsetof(PlyRangePt, pindex);
   range_props[nrp].count_offset = offsetof(PlyRangePt, npts);  
   nrp++;

   return nrp;
}


//
// Read in data from the ply file
//

int
readMeshPly(Mesh *mesh, char *filename)
{
   int i, j;
   int nelems;
   char **elist;
   int file_type;
   float version;
   char *elem_name;
   int nprops, num_vert_props, num_face_props;
   int num_elems;
   PlyProperty **plist;
   Vertex *vert;
   Triangle *tri;
   PlyVertex plyVert;
   PlyFace plyFace;
    
   PlyFile *ply = 
      ply_open_for_reading(filename, &nelems, &elist, &file_type, &version);
   
   if (!ply)
      return 0;

   whichVertProps(ply, mesh->hasNormals, mesh->hasIntensity, 
		  mesh->hasConfidence, mesh->hasColor); 
   
   num_vert_props = setupVertProps
      (ply, vert_props, mesh->hasNormals, mesh->hasIntensity, 
       mesh->hasConfidence, mesh->hasColor);
   
   num_face_props = setupFaceProps(ply, face_props);
   
   for (i = 0; i < nelems; i++) {
      
      /* get the description of the first element */
      elem_name = elist[i];
      plist = ply_get_element_description 
	 (ply, elem_name, &num_elems, &nprops);
      
      /* if we're on vertex elements, read them in */
      if (equal_strings ("vertex", elem_name)) {
	 
	 mesh->numVerts = num_elems;
	 mesh->verts = new Vertex[mesh->numVerts];
	 
	 /* set up for getting vertex elements */
	 ply_get_element_setup (ply, elem_name, num_vert_props, vert_props);
	 
	 /* grab all the vertex elements */
	 for (j = 0; j < mesh->numVerts; j++) {
	    ply_get_element (ply, (void *) &plyVert);
	    vert = &mesh->verts[j];
	    vert->coord.x = plyVert.x;
	    vert->coord.y = plyVert.y;
	    vert->coord.z = plyVert.z;
	    
	    if (mesh->hasConfidence)
	       vert->confidence = plyVert.confidence;
	    else 
	       vert->confidence = 1.0;
	 }
	 fflush(stdout);
	 
      }
      
      if (equal_strings ("face", elem_name)) {
	 
	 ply_get_element_setup (ply, elem_name, num_face_props, face_props);
	 
	 mesh->numTris = num_elems;
	 
	 if (mesh->numTris == 0)
	    continue;
	 
	 mesh->tris = new Triangle[mesh->numTris];
	 
	 for (j = 0; j < mesh->numTris; j++) {		
	    ply_get_element (ply, (void *) &plyFace);
	    tri = &mesh->tris[j];
	    tri->vindex1 = plyFace.vindex[0];
	    tri->vindex2 = plyFace.vindex[1];
	    tri->vindex3 = plyFace.vindex[2];
	    free(plyFace.vindex);
	 }
      }
   }
   
   return 1;
}


int
writeMeshPly(Mesh *mesh, char *filename)
{
   int i;
   float version;
   int num_vert_props, num_face_props;
   char *elem_names[] = {"vertex", "face"};
   
   PlyFile *ply = ply_open_for_writing(filename, 2, elem_names, 
				       PLY_ASCII, &version);
   
   if (ply == NULL)
      return 0;
   
   num_vert_props = setupVertProps
      (ply, vert_props, mesh->hasNormals, mesh->hasIntensity, 
       mesh->hasConfidence, mesh->hasColor);
   
   num_face_props = setupFaceProps(ply, face_props);

   ply_describe_element (ply, "vertex", mesh->numVerts, 
			 num_vert_props, vert_props);
   
   ply_describe_element (ply, "face", mesh->numTris, 
			 num_face_props, face_props);
   
   ply_header_complete (ply);
   
   /* set up and write the vertex elements */
   PlyVertex plyVert;
   Vertex *vert;
   
   ply_put_element_setup (ply, "vertex");
   
   for (i = 0; i < mesh->numVerts; i++) {
      vert = &mesh->verts[i];
      plyVert.x = vert->coord.x;
      plyVert.y = vert->coord.y;
      plyVert.z = vert->coord.z;

      if (mesh->hasConfidence) {
	 plyVert.confidence = vert->confidence;
      }

      if (mesh->hasNormals) {
	 plyVert.nx = vert->norm.x;
	 plyVert.ny = vert->norm.y;
	 plyVert.nz = vert->norm.z;
      }

      ply_put_element (ply, (void *) &plyVert);
   }
   
   PlyFace plyFace;
   Triangle *tri;
   int vertIndices[3];
   
   ply_put_element_setup (ply, "face");
   
   for (i = 0; i < mesh->numTris; i++) {
      tri = &mesh->tris[i];
      plyFace.nverts = 3;
      vertIndices[0] = tri->vindex1;
      vertIndices[1] = tri->vindex2;
      vertIndices[2] = tri->vindex3;
      plyFace.vindex = (int *)vertIndices;
      
      ply_put_element (ply, (void *) &plyFace);
   }
   
   /* close the PLY file */
   ply_close (ply);    
   
   return 1;
}



/******************************************************************************
									       Read range data from a PLY file.
									       
									       Entry:
  name - name of PLY file to read from

Exit:
  returns pointer to data, or NULL if it couldn't read from file
******************************************************************************/

int
readRangeGridPly(RangeGrid *rangeGrid, char *filename)
{
   int i, j, k, best_index, index;
   PlyFile *ply;
   PlyRangePt plyRangePt;
   PlyVertex plyVert;
   char **obj_info;
   int num_obj_info, num_vert_props, num_elems, num_range_props;
   int nelems, nprops;
   char **elist;
   int file_type;
   float version;
   PlyProperty **plist;
   int num_rows,num_cols;
   char *elem_name;
   char temp[PATH_MAX];
   float max;
   
   ply = ply_open_for_reading(filename, &nelems, &elist, &file_type, &version);

   if (ply == NULL)
      return 0;
   
   // see if we've got both vertex and range_grid data
   plist = ply_get_element_description (ply, "vertex", &num_elems, &nprops);
   if (plist == NULL) {
      fprintf (stderr, "file doesn't contain vertex data\n");
      return 0;
   }
   rangeGrid->numSamples = num_elems;

   plist = ply_get_element_description 
      (ply, "range_grid", &num_elems, &nprops);
   if (plist == NULL) {
      fprintf (stderr, "file doesn't contain range_grid data\n");
      return 0;
   }

   /* parse the obj_info */
   
   obj_info = ply_get_obj_info (ply, &num_obj_info);
   for (i = 0; i < num_obj_info; i++) {
      if (strstr(obj_info[i], "num_cols"))
	 sscanf(obj_info[i], "%s%d", temp, &num_cols);
      if (strstr(obj_info[i], "num_rows"))
	 sscanf(obj_info[i], "%s%d", temp, &num_rows);
   }

   /* set up the range data structure */
   rangeGrid->nlg = num_rows;
   rangeGrid->nlt = num_cols;

   rangeGrid->num_obj_info = num_obj_info;
   rangeGrid->obj_info = new char*[num_obj_info];
   for (i = 0; i < num_obj_info; i++) {
      rangeGrid->obj_info[i] = new char[strlen(obj_info[i])+1];
      strcpy(rangeGrid->obj_info[i], obj_info[i]);
   }
   
   whichVertProps(ply, rangeGrid->hasNormals, rangeGrid->hasIntensity, 
		  rangeGrid->hasConfidence, rangeGrid->hasColor); 
   
   num_vert_props = setupVertProps
      (ply, vert_props, rangeGrid->hasNormals, rangeGrid->hasIntensity, 
       rangeGrid->hasConfidence, rangeGrid->hasColor);
   
   rangeGrid->coords = new Vec3f[rangeGrid->numSamples];

   if (rangeGrid->hasConfidence)
      rangeGrid->confidence = new float[rangeGrid->numSamples];
   
   if (rangeGrid->hasIntensity)
      rangeGrid->intensity = new float[rangeGrid->numSamples];
   
   if (rangeGrid->hasColor)
      rangeGrid->matDiff = new vec3uc[rangeGrid->numSamples];
   
   num_range_props = setupRangeProps(ply, range_props);

   rangeGrid->indices = new int[rangeGrid->nlt * rangeGrid->nlg];


   /* read in the range data */
   
   for (i = 0; i < nelems; i++) {
      
      elem_name = elist[i];
      plist = ply_get_element_description 
	 (ply, elem_name, &num_elems, &nprops);
      
      if (equal_strings ("vertex", elem_name)) {
	 
	 /* set up for getting vertex elements */
	 ply_get_element_setup (ply, elem_name, num_vert_props, vert_props);

	 for (j = 0; j < num_elems; j++) {
	    ply_get_element (ply, (void *) &plyVert);
	    rangeGrid->coords[j].x = plyVert.x;
	    rangeGrid->coords[j].y = plyVert.y;
	    rangeGrid->coords[j].z = plyVert.z;
	    
	    if (rangeGrid->hasIntensity) {
	       rangeGrid->intensity[j] = plyVert.intensity;
	    }
	    
	    if (rangeGrid->hasConfidence) {
	       rangeGrid->confidence[j] = plyVert.confidence;
	    }
	    
	    if (rangeGrid->hasColor) {
	       rangeGrid->matDiff[j][0] = plyVert.diff_r;
	       rangeGrid->matDiff[j][1] = plyVert.diff_g;
	       rangeGrid->matDiff[j][2] = plyVert.diff_b;
	    }

	    if (rangeGrid->hasNormals) {
	       rangeGrid->norms[j].x = plyVert.nx;
	       rangeGrid->norms[j].y = plyVert.ny;
	       rangeGrid->norms[j].z = plyVert.nz;
	    }
	 }
      }
      
      if (equal_strings ("range_grid", elem_name)) {

	 ply_get_element_setup (ply, elem_name, num_range_props, range_props);

	 for (j = 0; j < num_elems; j++) {
	    ply_get_element (ply, (void *) &plyRangePt);
	    if (plyRangePt.npts == 0)
	       rangeGrid->indices[j] = -1;
	    else {
	       // Assumption: if there are multiple points, take the
	       // first one or the one with the brightest intensity,
	       // if provided
	       if (rangeGrid->hasIntensity) {
		  max = -FLT_MAX;
		  for (k = 0; k < plyRangePt.npts; k++) {
		     index = plyRangePt.pindex[k];		     
		     if (rangeGrid->intensity[index] > max) {
			max = rangeGrid->intensity[index];
			best_index = index;
		     }
		  }
	       }
	       else {
		  best_index = plyRangePt.pindex[0];
	       }
	       rangeGrid->indices[j] = best_index;
	       
	       free (plyRangePt.pindex);

	       fflush(stderr);
	    }
	 }
      }
   }
   
   ply_close (ply);
   
   return 1;
}

    
int
writeRangeGridPly(RangeGrid *rangeGrid, char *filename)
{
   int i, index;
   float version;
   int num_vert_props, num_range_props;
   char *elem_names[] = {"vertex", "range_grid"};
   
   PlyFile *ply = ply_open_for_writing(filename, 2, elem_names, 
				       PLY_ASCII, &version);
   
   if (ply == NULL)
      return 0;

   char obj_info[PATH_MAX];
   sprintf(obj_info, "%s %d", "num_cols", rangeGrid->nlt);   
   ply_put_obj_info (ply, obj_info);
   sprintf(obj_info, "%s %d", "num_rows", rangeGrid->nlg);   
   ply_put_obj_info (ply, obj_info);

   
   num_vert_props = setupVertProps
      (ply, vert_props, rangeGrid->hasNormals, rangeGrid->hasIntensity, 
       rangeGrid->hasConfidence, rangeGrid->hasColor);
   
   num_range_props = setupRangeProps(ply, range_props);

   ply_describe_element (ply, "vertex", rangeGrid->numSamples, 
			 num_vert_props, vert_props);
   
   ply_describe_element (ply, "range_grid", rangeGrid->nlg*rangeGrid->nlt, 
			 num_range_props, range_props);
   
   ply_header_complete (ply);
   
   /* set up and write the vertex elements */
   PlyVertex plyVert;
   
   ply_put_element_setup (ply, "vertex");
   
   for (i = 0; i < rangeGrid->numSamples; i++) {
      plyVert.x = rangeGrid->coords[i].x;
      plyVert.y = rangeGrid->coords[i].y;
      plyVert.z = rangeGrid->coords[i].z;

      if (rangeGrid->hasConfidence) {
	 plyVert.confidence = rangeGrid->confidence[i];
      }

      if (rangeGrid->hasIntensity) {
	 plyVert.intensity = rangeGrid->intensity[i];
      }

      if (rangeGrid->hasColor) {
	 plyVert.diff_r = rangeGrid->matDiff[i][0];
	 plyVert.diff_g = rangeGrid->matDiff[i][1];
	 plyVert.diff_b = rangeGrid->matDiff[i][2];
      }

      if (rangeGrid->hasNormals) {
	 plyVert.nx = rangeGrid->norms[i].x;
	 plyVert.ny = rangeGrid->norms[i].y;
	 plyVert.nz = rangeGrid->norms[i].z;
      }

      ply_put_element (ply, (void *) &plyVert);
   }
   
   PlyRangePt plyRangePt;
   
   ply_put_element_setup (ply, "range_grid");

   for (i = 0; i < rangeGrid->nlg*rangeGrid->nlt; i++) {
      index = rangeGrid->indices[i];
      if (index < 0) {
	 plyRangePt.npts = 0;
      }
      else {
	 plyRangePt.npts = 1;
	 plyRangePt.pindex = (int *) &index;
      }
      
      ply_put_element (ply, (void *) &plyRangePt);
   }
   
   /* close the PLY file */
   ply_close (ply);    
   
   return 1;
}


int 
isRangeGridFile(char *filename)
{
   int i;
   PlyFile *ply;
   int nelems;
   char **elist;
   int file_type;
   float version;
   
   ply = ply_open_for_reading(filename, &nelems, &elist, &file_type, &version);
   if (ply == NULL)
      return 0;
   ply_close(ply);
   
   for (i = 0; i < nelems; i++) {
      if (!strcmp(elist[i], "range_grid"))
	 return 1;
   }
   
   return 0;
}


